/*
 * Copyright (C) 2018-2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
/**
 *  @file
 *  create nsset
 */

#ifndef CREATE_NSSET_HH_85F3D80AEDC6465FB71D1306CCEADAEE
#define CREATE_NSSET_HH_85F3D80AEDC6465FB71D1306CCEADAEE

#include "libfred/opexception.hh"
#include "libfred/object/object.hh"
#include "libfred/opcontext.hh"
#include "util/optional_value.hh"
#include "util/db/nullable.hh"
#include "util/printable.hh"

#include "libfred/registrable_object/nsset/nsset_dns_host.hh"

#include <boost/date_time/posix_time/posix_time.hpp>

#include <string>
#include <vector>

namespace LibFred {

/**
* Create of nsset.
* Created instance is modifiable by chainable methods i.e. methods returning instance reference.
* Data set into instance by constructor and methods serve as input data of the create.
* Create is executed by @ref exec method with database connection supplied in @ref OperationContext parameter.
* In case of wrong input data or other predictable and superable failure, an instance of @ref CreateNsset::Exception is thrown with appropriate attributes set.
* In case of other insuperable failures and inconsistencies, an instance of @ref InternalError or other exception is thrown.
*/
class CreateNsset : public Util::Printable<CreateNsset>
{
public:
    DECLARE_EXCEPTION_DATA(invalid_nsset_handle, std::string);/**< exception members for invalid nsset handle generated by macro @ref DECLARE_EXCEPTION_DATA*/
    DECLARE_VECTOR_OF_EXCEPTION_DATA(already_set_dns_host, std::string);/**< exception members for vector of duplicated DNS hosts in nsset generated by macro @ref DECLARE_VECTOR_OF_EXCEPTION_DATA*/
    DECLARE_VECTOR_OF_EXCEPTION_DATA(invalid_dns_host_ipaddr, std::string);/**< exception members for vector of invalid DNS hosts in nsset generated by macro @ref DECLARE_VECTOR_OF_EXCEPTION_DATA*/
    DECLARE_VECTOR_OF_EXCEPTION_DATA(unknown_technical_contact_handle, std::string);/**< exception members for vector of unknown technical contact handles generated by macro @ref DECLARE_VECTOR_OF_EXCEPTION_DATA*/
    DECLARE_VECTOR_OF_EXCEPTION_DATA(already_set_technical_contact_handle, std::string);/**< exception members for vector of already set technical contact handles generated by macro @ref DECLARE_VECTOR_OF_EXCEPTION_DATA*/
    DECLARE_EXCEPTION_DATA(unknown_registrar_handle, std::string);/**< exception members for unknown registrar generated by macro @ref DECLARE_EXCEPTION_DATA*/

    struct Exception
    : virtual LibFred::OperationException
    , ExceptionData_invalid_nsset_handle<Exception>
    , ExceptionData_vector_of_unknown_technical_contact_handle<Exception>
    , ExceptionData_vector_of_already_set_technical_contact_handle<Exception>
    , ExceptionData_vector_of_already_set_dns_host<Exception>
    , ExceptionData_vector_of_invalid_dns_host_ipaddr<Exception>
    , ExceptionData_unknown_registrar_handle<Exception>
    {};

    /**
    * Create nsset constructor with mandatory parameters.
    * @param handle sets nsset handle into @ref handle_ attribute
    * @param registrar sets registrar handle into @ref registrar_ attribute
    */
    CreateNsset(const std::string& handle, const std::string& registrar);
    /**
    * Create nsset constructor with all parameters.
    * @param handle sets nsset handle into @ref handle_ attribute
    * @param registrar sets registrar handle into @ref registrar_ attribute
    * @param authinfo sets transfer password into @ref authinfo_ attribute
    * @param tech_check_level sets level of technical checks into @ref tech_check_level_ attribute
    * @param dns_hosts sets DNS hosts into @ref dns_hosts_ attribute
    * @param tech_contact sets list of technical contact handles into @ref tech_contacts_ attribute
    * @param logd_request_id sets logger request id into @ref logd_request_id_ attribute
    */
    CreateNsset(const std::string& handle
            , const std::string& registrar
            , const Optional<std::string>& authinfo
            , const Optional<short>& tech_check_level
            , const std::vector<DnsHost>& dns_hosts
            , const std::vector<std::string>& tech_contacts
            , const Optional<unsigned long long>& logd_request_id
            );

    /**
    * Sets nsset transfer password.
    * @param authinfo sets transfer password into @ref authinfo_ attribute
    * @return operation instance reference to allow method chaining
    */
    CreateNsset& set_authinfo(const std::string& authinfo);

    /**
    * Sets nsset level of technical checks.
    * @param tech_check_level sets level of technical checks into @ref tech_check_level_ attribute
    * @return operation instance reference to allow method chaining
    */
    CreateNsset& set_tech_check_level(short tech_check_level);

    /**
    * Sets nsset DNS hosts.
    * @param dns_hosts sets list of DNS hosts into @ref dns_hosts_ attribute
    * @return operation instance reference to allow method chaining
    */
    CreateNsset& set_dns_hosts(const std::vector<DnsHost>& dns_hosts);

    /**
    * Sets nsset technical contacts.
    * @param tech_contacts sets list of technical contact handles into @ref tech_contacts_ attribute
    * @return operation instance reference to allow method chaining
    */
    CreateNsset& set_tech_contacts(const std::vector<std::string>& tech_contacts);

    /**
    * Sets logger request id
    * @param logd_request_id sets logger request id into @ref logd_request_id_ attribute
    * @return operation instance reference to allow method chaining
    */
    CreateNsset& set_logd_request_id(unsigned long long logd_request_id);

    struct Result
    {
        CreateObject::Result create_object_result;
        boost::posix_time::ptime creation_time;///< timestamp of the contact creation
    };

    /**
    * Executes create
    * @param ctx contains reference to database and logging interface
    * @param returned_timestamp_pg_time_zone_name is postgresql time zone name of the returned timestamp
    * @return object id, history id and creation timestamp
    */
    Result exec(OperationContext& ctx, const std::string& returned_timestamp_pg_time_zone_name = "Europe/Prague");

    /**
    * Dumps state of the instance into the string
    * @return string with description of the instance state
    */
    std::string to_string()const;
protected:
    const std::string& get_handle()const;
private:
    const std::string handle_;/**< nsset identifier */
    const std::string registrar_;/**< handle of registrar performing the create */
    Optional<std::string> authinfo_;/**< transfer password */
    Optional<short> tech_check_level_; /**< level of technical checks*/
    std::vector<DnsHost> dns_hosts_; /**< DNS hosts of the nsset */
    std::vector<std::string> tech_contacts_; /**< technical contact handles */
    Nullable<unsigned long long> logd_request_id_; /**< id of the record in logger database, id is used in other calls to logging within current request */
};

}//namespace LibFred

#endif//CREATE_NSSET_HH_85F3D80AEDC6465FB71D1306CCEADAEE
