/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/diagnostics/about.hh"

#include "config.h"

#include <regex>
#include <stdexcept>
#include <string>

namespace Fred {
namespace Registry {
namespace Diagnostics {

AboutReply::Version AboutReply::Version::from_string(const std::string& full)
{
    Version version;
    //                                      1          2          3       4             6
    //                                      |------|   |------|   |------||-----------| |-------------------------|
    //                                      |      |   |      |   |      ||   5       | | 7         8             |
    //                                      |      |   |      |   |      ||   |------|| | |------|  |------------||
    static const std::regex version_regex("^([0-9]+)\\.([0-9]+)\\.([0-9]+)(-rc([0-9]+))?(-([0-9]+)-g([0-9a-fA-F]+))?$",
                                          std::regex_constants::egrep |
                                          std::regex_constants::optimize);
    std::smatch matches;
    if (!std::regex_match(full, matches, version_regex))
    {
        struct InvalidVersionString : std::runtime_error
        {
            InvalidVersionString() : std::runtime_error("invalid version string") { }
        };
        throw InvalidVersionString();
    }

    if (matches.size() != 9)
    {
        struct UnexpectedNumberOfMatches : std::runtime_error
        {
            UnexpectedNumberOfMatches() : std::runtime_error("unexpected number of matches") { }
        };
        throw UnexpectedNumberOfMatches();
    }
    version.full = full;

    if (!matches[1].matched)
    {
        struct NoMajorVersionFound : std::runtime_error
        {
            NoMajorVersionFound() : std::runtime_error("no major version found") { }
        };
        throw NoMajorVersionFound();
    }
    version.major = std::stol(matches[1]);

    if (!matches[2].matched)
    {
        struct NoMinorVersionFound : std::runtime_error
        {
            NoMinorVersionFound() : std::runtime_error("no minor version found") { }
        };
        throw NoMinorVersionFound();
    }
    version.minor = std::stol(matches[2]);

    if (!matches[3].matched)
    {
        struct NoPatchVersionFound : std::runtime_error
        {
            NoPatchVersionFound() : std::runtime_error("no patch version found") { }
        };
        throw NoPatchVersionFound();
    }
    version.patch = std::stol(matches[3]);

    if (matches[4].matched)
    {
        if (!matches[5].matched)
        {
            struct NoReleaseCandidateVersionFound : std::runtime_error
            {
                NoReleaseCandidateVersionFound() : std::runtime_error("no release candidate version found") { }
            };
            throw NoReleaseCandidateVersionFound();
        }
        version.rc = std::stol(matches[5]);
    }
    if (matches[6].matched)
    {
        Version::Commit commit;
        if (!matches[7].matched)
        {
            struct NoCommitDistanceFound : std::runtime_error
            {
                NoCommitDistanceFound() : std::runtime_error("no commit distance found") { }
            };
            throw NoCommitDistanceFound();
        }
        commit.distance = std::stol(matches[7]);

        if (!matches[8].matched)
        {
            struct NoCommitHashFound : std::runtime_error
            {
                NoCommitHashFound() : std::runtime_error("no commit hash found") { }
            };
            throw NoCommitHashFound();
        }
        commit.hash = matches[8];

        version.commit = commit;
    }
    return version;
}

AboutReply::Data about()
{
    AboutReply::Data reply;
    reply.server_version = AboutReply::Version::from_string(SERVER_VERSION);
    reply.api_version = AboutReply::Version::from_string(API_VERSION);
    return reply;
}

}//namespace Fred::Registry::Diagnostics
}//namespace Fred::Registry
}//namespace Fred
