/*
 * Copyright (C) 2019-2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/proto_unwrap.hh"
#include "src/util/into.hh"

#include "util/log/logger.hh"

#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

namespace Fred {
namespace Registry {

template <>
struct ProtoTypeTraits<std::string>
{
    using RegistryType = std::string;
};

template <typename T>
typename ProtoTypeTraits<T>::RegistryType proto_unwrap(const T& src)
{
    static_assert(std::is_same<typename ProtoTypeTraits<T>::RegistryType, T>::value, "no conversion available");
    return src;
}

template <>
struct ProtoTypeTraits<Registry::Api::Uuid>
{
    using RegistryType = boost::uuids::uuid;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactId>
{
    using RegistryType = Contact::ContactId;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactHistoryId>
{
    using RegistryType = Contact::ContactHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactHandle>
{
    using RegistryType = Contact::ContactHandle;
};

template <>
struct ProtoTypeTraits<Api::Contact::ContactHistoryInterval>
{
    using RegistryType = Contact::ContactHistoryInterval;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainId>
{
    using RegistryType = Domain::DomainId;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainHistoryId>
{
    using RegistryType = Domain::DomainHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Domain::Fqdn>
{
    using RegistryType = Domain::Fqdn;
};

template <>
struct ProtoTypeTraits<Api::Domain::DomainHistoryInterval>
{
    using RegistryType = Domain::DomainHistoryInterval;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetId>
{
    using RegistryType = Keyset::KeysetId;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetHistoryId>
{
    using RegistryType = Keyset::KeysetHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetHandle>
{
    using RegistryType = Keyset::KeysetHandle;
};

template <>
struct ProtoTypeTraits<Api::Keyset::KeysetHistoryInterval>
{
    using RegistryType = Keyset::KeysetHistoryInterval;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetId>
{
    using RegistryType = Nsset::NssetId;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetHistoryId>
{
    using RegistryType = Nsset::NssetHistoryId;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetHandle>
{
    using RegistryType = Nsset::NssetHandle;
};

template <>
struct ProtoTypeTraits<Api::Nsset::NssetHistoryInterval>
{
    using RegistryType = Nsset::NssetHistoryInterval;
};

template <>
boost::uuids::uuid proto_unwrap(const Registry::Api::Uuid& src)
{
    return boost::uuids::string_generator()(src.value());
}

template <>
Contact::ContactId proto_unwrap(const Api::Contact::ContactId& src)
{
    return Util::make_strong<Contact::ContactId>(proto_unwrap(src.uuid()));
}

template <>
Contact::ContactHistoryId proto_unwrap(const Api::Contact::ContactHistoryId& src)
{
    return Util::make_strong<Contact::ContactHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Diagnostics::PingRequest proto_unwrap(const Api::Diagnostics::PingRequest& src)
{
    Diagnostics::PingRequest dst;
    dst.check_database_accessibility = src.check_database_accessibility();
    return dst;
}

template <>
Contact::ContactInfoRequest proto_unwrap(const Api::Contact::ContactInfoRequest& src)
{
    Contact::ContactInfoRequest dst;
    dst.contact_id = proto_unwrap(src.contact_id());
    if (src.has_contact_history_id())
    {
        dst.contact_history_id = proto_unwrap(src.contact_history_id());
    }
    return dst;
}

template <>
Contact::ContactHandle proto_unwrap(const Api::Contact::ContactHandle& src)
{
    return Util::make_strong<Contact::ContactHandle>(proto_unwrap(src.value()));
}

namespace {

Contact::ContactHistoryInterval::HistoryId unwrap_contact_history_interval_history_id(const Api::Contact::ContactHistoryId& src)
{
    return Util::make_strong<Contact::ContactHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

std::chrono::time_point<std::chrono::system_clock, std::chrono::nanoseconds> unwrap_system_clock_nanoseconds(
        const google::protobuf::Timestamp& src)
{
    const std::chrono::time_point<std::chrono::system_clock, std::chrono::nanoseconds> dst(
            std::chrono::nanoseconds(1'000'000'000 * src.seconds() + src.nanos()));
    return dst;
}

}//namespace Fred::Registry::{anonymous}

template <>
Contact::ContactHistoryInterval proto_unwrap(const Api::Contact::ContactHistoryInterval& src)
{
    auto lower_limit = Contact::ContactHistoryInterval::LowerLimit(Contact::ContactHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_contact_history_id())
        {
            lower_limit.value = unwrap_contact_history_interval_history_id(src.lower_limit().contact_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Contact::ContactHistoryInterval::UpperLimit(Contact::ContactHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_contact_history_id())
        {
            upper_limit.value = unwrap_contact_history_interval_history_id(src.upper_limit().contact_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Contact::ContactHistoryInterval(lower_limit, upper_limit);
}

template <>
Contact::SearchContactRequest proto_unwrap(const Api::Contact::SearchContactRequest& src)
{
    Contact::SearchContactRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Contact::SearchContactRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Contact::ContactItem::contact_handle))
        {
            dst.searched_items.insert(Contact::ContactItem::contact_handle);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::name))
        {
            dst.searched_items.insert(Contact::ContactItem::name);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::organization))
        {
            dst.searched_items.insert(Contact::ContactItem::organization);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_street))
        {
            dst.searched_items.insert(Contact::ContactItem::place_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_city))
        {
            dst.searched_items.insert(Contact::ContactItem::place_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::telephone))
        {
            dst.searched_items.insert(Contact::ContactItem::telephone);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::fax))
        {
            dst.searched_items.insert(Contact::ContactItem::fax);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::email))
        {
            dst.searched_items.insert(Contact::ContactItem::email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::notify_email))
        {
            dst.searched_items.insert(Contact::ContactItem::notify_email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::vat_identification_number))
        {
            dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::additional_identifier))
        {
            dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_company))
        {
            dst.searched_items.insert(Contact::ContactItem::address_company);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_street))
        {
            dst.searched_items.insert(Contact::ContactItem::address_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_city))
        {
            dst.searched_items.insert(Contact::ContactItem::address_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::address_postal_code);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchContactRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Contact::ContactItem::contact_handle);
        dst.searched_items.insert(Contact::ContactItem::name);
        dst.searched_items.insert(Contact::ContactItem::organization);
        dst.searched_items.insert(Contact::ContactItem::place_street);
        dst.searched_items.insert(Contact::ContactItem::place_city);
        dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        dst.searched_items.insert(Contact::ContactItem::telephone);
        dst.searched_items.insert(Contact::ContactItem::fax);
        dst.searched_items.insert(Contact::ContactItem::email);
        dst.searched_items.insert(Contact::ContactItem::notify_email);
        dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        dst.searched_items.insert(Contact::ContactItem::address_company);
        dst.searched_items.insert(Contact::ContactItem::address_street);
        dst.searched_items.insert(Contact::ContactItem::address_city);
        dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::address_postal_code);
    }
    return dst;
}

template <>
Contact::SearchContactHistoryRequest proto_unwrap(const Api::Contact::SearchContactHistoryRequest& src)
{
    Contact::SearchContactHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Contact::SearchContactHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_contacts = src.max_number_of_contacts();
    }
    else
    {
        dst.max_number_of_contacts = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Contact::ContactItem::contact_handle))
        {
            dst.searched_items.insert(Contact::ContactItem::contact_handle);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::name))
        {
            dst.searched_items.insert(Contact::ContactItem::name);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::organization))
        {
            dst.searched_items.insert(Contact::ContactItem::organization);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_street))
        {
            dst.searched_items.insert(Contact::ContactItem::place_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_city))
        {
            dst.searched_items.insert(Contact::ContactItem::place_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::place_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::telephone))
        {
            dst.searched_items.insert(Contact::ContactItem::telephone);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::fax))
        {
            dst.searched_items.insert(Contact::ContactItem::fax);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::email))
        {
            dst.searched_items.insert(Contact::ContactItem::email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::notify_email))
        {
            dst.searched_items.insert(Contact::ContactItem::notify_email);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::vat_identification_number))
        {
            dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::additional_identifier))
        {
            dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_company))
        {
            dst.searched_items.insert(Contact::ContactItem::address_company);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_street))
        {
            dst.searched_items.insert(Contact::ContactItem::address_street);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_city))
        {
            dst.searched_items.insert(Contact::ContactItem::address_city);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_state_or_province))
        {
            dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        }
        else if (item == Util::Into<std::string>::from(Contact::ContactItem::address_postal_code))
        {
            dst.searched_items.insert(Contact::ContactItem::address_postal_code);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchContactHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Contact::ContactItem::contact_handle);
        dst.searched_items.insert(Contact::ContactItem::name);
        dst.searched_items.insert(Contact::ContactItem::organization);
        dst.searched_items.insert(Contact::ContactItem::place_street);
        dst.searched_items.insert(Contact::ContactItem::place_city);
        dst.searched_items.insert(Contact::ContactItem::place_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::place_postal_code);
        dst.searched_items.insert(Contact::ContactItem::telephone);
        dst.searched_items.insert(Contact::ContactItem::fax);
        dst.searched_items.insert(Contact::ContactItem::email);
        dst.searched_items.insert(Contact::ContactItem::notify_email);
        dst.searched_items.insert(Contact::ContactItem::vat_identification_number);
        dst.searched_items.insert(Contact::ContactItem::additional_identifier);
        dst.searched_items.insert(Contact::ContactItem::address_company);
        dst.searched_items.insert(Contact::ContactItem::address_street);
        dst.searched_items.insert(Contact::ContactItem::address_city);
        dst.searched_items.insert(Contact::ContactItem::address_state_or_province);
        dst.searched_items.insert(Contact::ContactItem::address_postal_code);
    }
    return dst;
}

template <>
Contact::ContactDataHistoryRequest proto_unwrap(const Api::Contact::ContactHistoryRequest& src)
{
    return Contact::ContactDataHistoryRequest(proto_unwrap(src.contact_id()),
                                              proto_unwrap(src.history()));
}

template <>
Contact::ContactHandleHistoryRequest proto_unwrap(const Api::Contact::ContactHandleHistoryRequest& src)
{
    return Contact::ContactHandleHistoryRequest(proto_unwrap(src.contact_handle()));
}

template <>
Contact::ContactStateRequest proto_unwrap(const Api::Contact::ContactStateRequest& src)
{
    Contact::ContactStateRequest dst;
    dst.contact_id = proto_unwrap(src.contact_id());
    return dst;
}

template <>
Contact::ContactStateHistoryRequest proto_unwrap(const Api::Contact::ContactStateHistoryRequest& src)
{
    return Contact::ContactStateHistoryRequest(proto_unwrap(src.contact_id()), proto_unwrap(src.history()));
}

template <>
Keyset::KeysetId proto_unwrap(const Api::Keyset::KeysetId& src)
{
    return Util::make_strong<Keyset::KeysetId>(proto_unwrap(src.uuid()));
}

template <>
Keyset::KeysetHistoryId proto_unwrap(const Api::Keyset::KeysetHistoryId& src)
{
    return Util::make_strong<Keyset::KeysetHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Keyset::KeysetInfoRequest proto_unwrap(const Api::Keyset::KeysetInfoRequest& src)
{
    Keyset::KeysetInfoRequest dst;
    dst.keyset_id = proto_unwrap(src.keyset_id());
    if (src.has_keyset_history_id())
    {
        dst.keyset_history_id = proto_unwrap(src.keyset_history_id());
    }
    return dst;
}

template <>
Keyset::KeysetHandle proto_unwrap(const Api::Keyset::KeysetHandle& src)
{
    return Util::make_strong<Keyset::KeysetHandle>(proto_unwrap(src.value()));
}

namespace {

Keyset::KeysetHistoryInterval::HistoryId unwrap_keyset_history_interval_history_id(const Api::Keyset::KeysetHistoryId& src)
{
    return Util::make_strong<Keyset::KeysetHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

}//namespace Fred::Registry::{anonymous}

template <>
Keyset::KeysetHistoryInterval proto_unwrap(const Api::Keyset::KeysetHistoryInterval& src)
{
    auto lower_limit = Keyset::KeysetHistoryInterval::LowerLimit(Keyset::KeysetHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_keyset_history_id())
        {
            lower_limit.value = unwrap_keyset_history_interval_history_id(src.lower_limit().keyset_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Keyset::KeysetHistoryInterval::UpperLimit(Keyset::KeysetHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_keyset_history_id())
        {
            upper_limit.value = unwrap_keyset_history_interval_history_id(src.upper_limit().keyset_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Keyset::KeysetHistoryInterval(lower_limit, upper_limit);
}

template <>
Keyset::SearchKeysetRequest proto_unwrap(const Api::Keyset::SearchKeysetRequest& src)
{
    Keyset::SearchKeysetRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Keyset::SearchKeysetRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Keyset::KeysetItem::keyset_handle))
        {
            dst.searched_items.insert(Keyset::KeysetItem::keyset_handle);
        }
        else if (item == Util::Into<std::string>::from(Keyset::KeysetItem::public_key))
        {
            dst.searched_items.insert(Keyset::KeysetItem::public_key);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchKeysetRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Keyset::KeysetItem::keyset_handle,
                               Keyset::KeysetItem::public_key };
    }
    return dst;
}

template <>
Keyset::SearchKeysetHistoryRequest proto_unwrap(const Api::Keyset::SearchKeysetHistoryRequest& src)
{
    Keyset::SearchKeysetHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Keyset::SearchKeysetHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_keysets = src.max_number_of_keysets();
    }
    else
    {
        dst.max_number_of_keysets = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Keyset::KeysetItem::keyset_handle))
        {
            dst.searched_items.insert(Keyset::KeysetItem::keyset_handle);
        }
        else if (item == Util::Into<std::string>::from(Keyset::KeysetItem::public_key))
        {
            dst.searched_items.insert(Keyset::KeysetItem::public_key);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchKeysetHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Keyset::KeysetItem::keyset_handle,
                               Keyset::KeysetItem::public_key };
    }
    return dst;
}

template <>
Keyset::KeysetDataHistoryRequest proto_unwrap(const Api::Keyset::KeysetHistoryRequest& src)
{
    return Keyset::KeysetDataHistoryRequest(
            proto_unwrap(src.keyset_id()),
            proto_unwrap(src.history()));
}

template <>
Keyset::KeysetHandleHistoryRequest proto_unwrap(const Api::Keyset::KeysetHandleHistoryRequest& src)
{
    return Keyset::KeysetHandleHistoryRequest(proto_unwrap(src.keyset_handle()));
}

template <>
Keyset::KeysetStateRequest proto_unwrap(const Api::Keyset::KeysetStateRequest& src)
{
    Keyset::KeysetStateRequest dst;
    dst.keyset_id = proto_unwrap(src.keyset_id());
    return dst;
}

template <>
Keyset::KeysetStateHistoryRequest proto_unwrap(const Api::Keyset::KeysetStateHistoryRequest& src)
{
    return Keyset::KeysetStateHistoryRequest(proto_unwrap(src.keyset_id()), proto_unwrap(src.history()));
}

template <>
Nsset::NssetId proto_unwrap(const Api::Nsset::NssetId& src)
{
    return Util::make_strong<Nsset::NssetId>(proto_unwrap(src.uuid()));
}

template <>
Nsset::NssetHistoryId proto_unwrap(const Api::Nsset::NssetHistoryId& src)
{
    return Util::make_strong<Nsset::NssetHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Nsset::NssetInfoRequest proto_unwrap(const Api::Nsset::NssetInfoRequest& src)
{
    Nsset::NssetInfoRequest dst;
    dst.nsset_id = proto_unwrap(src.nsset_id());
    if (src.has_nsset_history_id())
    {
        dst.nsset_history_id = proto_unwrap(src.nsset_history_id());
    }
    return dst;
}

template <>
Nsset::NssetHandle proto_unwrap(const Api::Nsset::NssetHandle& src)
{
    return Util::make_strong<Nsset::NssetHandle>(proto_unwrap(src.value()));
}

namespace {

Nsset::NssetHistoryInterval::HistoryId unwrap_nsset_history_interval_history_id(const Api::Nsset::NssetHistoryId& src)
{
    return Util::make_strong<Nsset::NssetHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

}//namespace Fred::Registry::{anonymous}

template <>
Nsset::NssetHistoryInterval proto_unwrap(const Api::Nsset::NssetHistoryInterval& src)
{
    auto lower_limit = Nsset::NssetHistoryInterval::LowerLimit(Nsset::NssetHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_nsset_history_id())
        {
            lower_limit.value = unwrap_nsset_history_interval_history_id(src.lower_limit().nsset_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Nsset::NssetHistoryInterval::UpperLimit(Nsset::NssetHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_nsset_history_id())
        {
            upper_limit.value = unwrap_nsset_history_interval_history_id(src.upper_limit().nsset_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Nsset::NssetHistoryInterval(lower_limit, upper_limit);
}

template <>
Nsset::SearchNssetRequest proto_unwrap(const Api::Nsset::SearchNssetRequest& src)
{
    Nsset::SearchNssetRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Nsset::SearchNssetRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Nsset::NssetItem::nsset_handle))
        {
            dst.searched_items.insert(Nsset::NssetItem::nsset_handle);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_fqdn))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_fqdn);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_ip_addresses))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_ip_addresses);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchNssetRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Nsset::NssetItem::nsset_handle,
                               Nsset::NssetItem::dns_hosts_fqdn,
                               Nsset::NssetItem::dns_hosts_ip_addresses };
    }
    return dst;
}

template <>
Nsset::SearchNssetHistoryRequest proto_unwrap(const Api::Nsset::SearchNssetHistoryRequest& src)
{
    Nsset::SearchNssetHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Nsset::SearchNssetHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_nssets = src.max_number_of_nssets();
    }
    else
    {
        dst.max_number_of_nssets = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Nsset::NssetItem::nsset_handle))
        {
            dst.searched_items.insert(Nsset::NssetItem::nsset_handle);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_fqdn))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_fqdn);
        }
        else if (item == Util::Into<std::string>::from(Nsset::NssetItem::dns_hosts_ip_addresses))
        {
            dst.searched_items.insert(Nsset::NssetItem::dns_hosts_ip_addresses);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchNssetHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items = { Nsset::NssetItem::nsset_handle,
                               Nsset::NssetItem::dns_hosts_fqdn,
                               Nsset::NssetItem::dns_hosts_ip_addresses };
    }
    return dst;
}

template <>
Nsset::NssetDataHistoryRequest proto_unwrap(const Api::Nsset::NssetHistoryRequest& src)
{
    return Nsset::NssetDataHistoryRequest(
            proto_unwrap(src.nsset_id()),
            proto_unwrap(src.history()));
}

template <>
Nsset::NssetHandleHistoryRequest proto_unwrap(const Api::Nsset::NssetHandleHistoryRequest& src)
{
    return Nsset::NssetHandleHistoryRequest(proto_unwrap(src.nsset_handle()));
}

template <>
Nsset::NssetStateRequest proto_unwrap(const Api::Nsset::NssetStateRequest& src)
{
    Nsset::NssetStateRequest dst;
    dst.nsset_id = proto_unwrap(src.nsset_id());
    return dst;
}

template <>
Nsset::NssetStateHistoryRequest proto_unwrap(const Api::Nsset::NssetStateHistoryRequest& src)
{
    return Nsset::NssetStateHistoryRequest(proto_unwrap(src.nsset_id()), proto_unwrap(src.history()));
}

template <>
Domain::DomainId proto_unwrap(const Api::Domain::DomainId& src)
{
    return Util::make_strong<Domain::DomainId>(proto_unwrap(src.uuid()));
}

template <>
Domain::DomainHistoryId proto_unwrap(const Api::Domain::DomainHistoryId& src)
{
    return Util::make_strong<Domain::DomainHistoryId>(proto_unwrap(src.uuid()));
}

template <>
Domain::DomainInfoRequest proto_unwrap(const Api::Domain::DomainInfoRequest& src)
{
    Domain::DomainInfoRequest dst;
    dst.domain_id = proto_unwrap(src.domain_id());
    if (src.has_domain_history_id())
    {
        dst.domain_history_id = proto_unwrap(src.domain_history_id());
    }
    return dst;
}

template <>
Domain::Fqdn proto_unwrap(const Api::Domain::Fqdn& src)
{
    return Util::make_strong<Domain::Fqdn>(proto_unwrap(src.value()));
}

namespace {

Domain::DomainHistoryInterval::HistoryId unwrap_domain_history_interval_history_id(const Api::Domain::DomainHistoryId& src)
{
    return Util::make_strong<Domain::DomainHistoryInterval::HistoryId>(
            Util::get_raw_value_from(proto_unwrap(src)));
}

}//namespace Fred::Registry::{anonymous}

template <>
Domain::DomainHistoryInterval proto_unwrap(const Api::Domain::DomainHistoryInterval& src)
{
    auto lower_limit = Domain::DomainHistoryInterval::LowerLimit(Domain::DomainHistoryInterval::NoLimit());
    if (src.has_lower_limit())
    {
        if (src.lower_limit().has_domain_history_id())
        {
            lower_limit.value = unwrap_domain_history_interval_history_id(src.lower_limit().domain_history_id());
        }
        else if (src.lower_limit().has_timestamp())
        {
            lower_limit.value = unwrap_system_clock_nanoseconds(src.lower_limit().timestamp());
        }
    }
    auto upper_limit = Domain::DomainHistoryInterval::UpperLimit(Domain::DomainHistoryInterval::NoLimit());
    if (src.has_upper_limit())
    {
        if (src.upper_limit().has_domain_history_id())
        {
            upper_limit.value = unwrap_domain_history_interval_history_id(src.upper_limit().domain_history_id());
        }
        else if (src.upper_limit().has_timestamp())
        {
            upper_limit.value = unwrap_system_clock_nanoseconds(src.upper_limit().timestamp());
        }
    }
    return Domain::DomainHistoryInterval(lower_limit, upper_limit);
}

template <>
Domain::SearchDomainRequest proto_unwrap(const Api::Domain::SearchDomainRequest& src)
{
    Domain::SearchDomainRequest dst;
    for (const auto& item : src.query_values())
    {
        dst.query_values.push_back(item);
    }

    if (src.Limit_case() != Api::Domain::SearchDomainRequest::LIMIT_NOT_SET)
    {
        dst.limit = src.limit();
    }
    else
    {
        dst.limit = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Domain::DomainItem::fqdn))
        {
            dst.searched_items.insert(Domain::DomainItem::fqdn);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchDomainRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Domain::DomainItem::fqdn);
    }
    return dst;
}

template <>
Domain::SearchDomainHistoryRequest proto_unwrap(const Api::Domain::SearchDomainHistoryRequest& src)
{
    Domain::SearchDomainHistoryRequest dst;
    dst.query_values.reserve(src.query_values_size());
    for (const auto& value : src.query_values())
    {
        dst.query_values.push_back(value);
    }
    if (src.Limit_case() != Api::Domain::SearchDomainHistoryRequest::LIMIT_NOT_SET)
    {
        dst.max_number_of_domains = src.max_number_of_domains();
    }
    else
    {
        dst.max_number_of_domains = boost::none;
    }
    if (src.has_data_valid_from())
    {
        dst.data_valid_from = unwrap_system_clock_nanoseconds(src.data_valid_from());
    }
    else
    {
        dst.data_valid_from = boost::none;
    }
    if (src.has_data_valid_to())
    {
        dst.data_valid_to = unwrap_system_clock_nanoseconds(src.data_valid_to());
    }
    else
    {
        dst.data_valid_to = boost::none;
    }
    for (const auto& item : src.searched_items())
    {
        if (item == Util::Into<std::string>::from(Domain::DomainItem::fqdn))
        {
            dst.searched_items.insert(Domain::DomainItem::fqdn);
        }
        else
        {
            LOGGER.warning("unknown item \"" + item + "\" in SearchDomainHistoryRequest");
        }
    }
    if (dst.searched_items.empty())
    {
        dst.searched_items.insert(Domain::DomainItem::fqdn);
    }
    return dst;
}

template <>
Domain::DomainDataHistoryRequest proto_unwrap(const Api::Domain::DomainHistoryRequest& src)
{
    return Domain::DomainDataHistoryRequest(
            proto_unwrap(src.domain_id()),
            proto_unwrap(src.history()));
}

template <>
Domain::FqdnHistoryRequest proto_unwrap(const Api::Domain::FqdnHistoryRequest& src)
{
    return Domain::FqdnHistoryRequest(proto_unwrap(src.fqdn()));
}

template <>
Domain::DomainStateRequest proto_unwrap(const Api::Domain::DomainStateRequest& src)
{
    Domain::DomainStateRequest dst;
    dst.domain_id = proto_unwrap(src.domain_id());
    return dst;
}

template <>
Domain::DomainStateHistoryRequest proto_unwrap(const Api::Domain::DomainStateHistoryRequest& src)
{
    return Domain::DomainStateHistoryRequest(proto_unwrap(src.domain_id()), proto_unwrap(src.history()));
}


template <>
struct ProtoTypeTraits<Api::Registrar::RegistrarHandle>
{
    using RegistryType = Registrar::RegistrarHandle;
};

template <>
Registrar::RegistrarHandle proto_unwrap(const Api::Registrar::RegistrarHandle& src)
{
    return Util::make_strong<Registrar::RegistrarHandle>(proto_unwrap(src.value()));
}

template <>
Registrar::RegistrarCreditRequest proto_unwrap(const Api::Registrar::RegistrarCreditRequest& src)
{
    Registrar::RegistrarCreditRequest dst;
    dst.registrar_handle = proto_unwrap(src.registrar_handle());
    return dst;
}

template <>
Registrar::RegistrarInfoRequest proto_unwrap(const Api::Registrar::RegistrarInfoRequest& src)
{
    Registrar::RegistrarInfoRequest dst;
    dst.registrar_handle = proto_unwrap(src.registrar_handle());
    return dst;
}

template <>
Registrar::RegistrarZoneAccessRequest proto_unwrap(const Api::Registrar::RegistrarZoneAccessRequest& src)
{
    Registrar::RegistrarZoneAccessRequest dst;
    dst.registrar_handle = proto_unwrap(src.registrar_handle());
    return dst;
}

}//namespace Fred::Registry
}//namespace Fred
